#include-once

; User Interface Bounding Box UDF

Global Const $UIBB_MAXITER = 10 ; Anzahl Fit-Iterationen. Je mehr, desto genauer werden Fixierte Größen (z.B. Breite = 100) eingehalten.

Global Const $UIBB_VBOX = -1 ; Vertical Vector of Elements
Global Const $UIBB_HBOX = -2 ; Horizontal Vector of Elements
Global Const $UIBB_EMPTY = -3 ; Padding


Func _UIBB_SetSize(ByRef $BOX, $nMinSizePx, $nMaxSizePx = -1)
	If $nMaxSizePx = -1 Or $nMaxSizePx > $nMinSizePx Then $nMaxSizePx = $nMinSizePx
	$BOX['MINSIZE'] = $nMinSizePx
	$BOX['MAXSIZE'] = $nMaxSizePx
EndFunc   ;==>_UIBB_SetSize

Func _UIBB_SetMargin(ByRef $BOX, $leftPx, $rightPx, $topPx, $botPx)
	$BOX['MARGIN'] = __UIBB_Margin($leftPx, $rightPx, $topPx, $botPx)
EndFunc   ;==>_UIBB_SetSize



Func _UIBB_Box($iTYPE = $UIBB_EMPTY, $hGUI = '')
	Local $BOX[], $MAP[]
	$BOX['TYPE'] = $iTYPE
	$BOX['RECT'] = __UIBB_Rect()
	$BOX['SCALE'] = 1
	$BOX['EFFSCALE'] = 1
	$BOX['MINSIZE'] = 0
	$BOX['MAXSIZE'] = 65535
	$BOX['NONBLOCKING'] = False
	$BOX['MARGIN'] = __UIBB_Margin()
	If IsPtr($hGUI) Then $BOX['GUI'] = $hGUI
	$BOX['CHILDREN'] = $MAP
	Return $BOX
EndFunc   ;==>_UIBB_Box

Func _UIBB_Insert(ByRef $PARENT, $xChild, $nScale = 1)
	If IsFunc($xChild) Then Return __UIBB_InsertCtrl($PARENT, $xChild, $nScale)
	Return __UIBB_InsertBox($PARENT, $xChild, $nScale)
EndFunc   ;==>_UIBB_Insert

Func _UIBB_Update(ByRef $BOX)
	__UIBB_UpdateClientArea($BOX, 0) ; Update everything using $hBOX as Root
EndFunc   ;==>_UIBB_Update



Func __UIBB_InsertCtrl(ByRef $PARENT, $xCtrl, $nScale)
	If Not __UIBB_IsBox($PARENT) Then
		ConsoleWrite('! ERROR: Cannot Insert Ctrl into another Ctrl. Insert it into a Box instead!' & @CRLF)
		Return SetError(1, 0, 0)
	EndIf
	Local $CHILD[]
	$CHILD['TYPE'] = $xCtrl
	$CHILD['RECT'] = __UIBB_Rect()
	If $nScale = -1 Then
		$CHILD['SCALE'] = 1
		$CHILD['EFFSCALE'] = 1
		$CHILD['NONBLOCKING'] = True
		ConsoleWrite('Nonblocking = true' & @CRLF)
	Else
		$CHILD['SCALE'] = $nScale
		$CHILD['EFFSCALE'] = $nScale
		$CHILD['NONBLOCKING'] = False
	EndIf
	$CHILD['MINSIZE'] = 0
	$CHILD['MAXSIZE'] = 65535
	Local $bSuccess = False
	Switch $xCtrl
		Case GUICtrlCreateAvi ; (hier muss eine Datei angegeben werden...)
		Case GUICtrlCreateButton, GUICtrlCreateCheckbox, GUICtrlCreateCombo, GUICtrlCreateDate, GUICtrlCreateEdit, GUICtrlCreateGroup, GUICtrlCreateInput, GUICtrlCreateLabel, GUICtrlCreateList, GUICtrlCreateListView, GUICtrlCreateRadio
			$CHILD['ID'] = $CHILD['TYPE']('', 0, 0)
			$bSuccess = True
		Case GUICtrlCreateDummy ; (das Ding hat keine Größe und kann demnach nicht automatisch skaliert werden)
		Case GUICtrlCreateGraphic
			$CHILD['ID'] = $CHILD['TYPE'](0, 0)
			$bSuccess = True
		Case GUICtrlCreateIcon ; (hier muss eine Datei angegeben werden...)
	EndSwitch
	If $bSuccess Then
		GUICtrlSetResizing(-1, 802) ; DOCKALL Wir machen den Resize ja von Hand.
		MapAppend($PARENT['CHILDREN'], $CHILD)
		Return $CHILD['ID']
	EndIf
	Return SetError(1, 0, 0) ; Da ist doch böses im Busch.
EndFunc   ;==>__UIBB_InsertCtrl

Func __UIBB_InsertBox(ByRef $PARENT, ByRef $CHILD, $nScale)
	If Not __UIBB_IsBox($PARENT) Then
		ConsoleWrite('! ERROR: Cannot Insert Box into another Ctrl. Insert it into a Box instead!' & @CRLF)
		Return SetError(1, 0, 0)
	EndIf
	If $nScale = -1 Then
		$CHILD['SCALE'] = 1
		$CHILD['EFFSCALE'] = 1
		$CHILD['NONBLOCKING'] = True
	Else
		$CHILD['SCALE'] = $nScale
		$CHILD['EFFSCALE'] = $nScale
		$CHILD['NONBLOCKING'] = False
	EndIf
	MapAppend($PARENT['CHILDREN'], $CHILD)
EndFunc   ;==>__UIBB_InsertBox

Func __UIBB_UpdateClientArea(ByRef $CHILD, ByRef $PARENT, $idx = -1)
	If __UIBB_IsBox($CHILD) Then
		__UIBB_UpdateBoxClientArea($CHILD, $PARENT, $idx)
	Else
		__UIBB_UpdateCtrlClientArea($CHILD, $PARENT, $idx)
	EndIf
	__UIBB_ComputeEffectiveScale($CHILD)
	For $i = 0 To UBound($CHILD['CHILDREN']) - 1 Step 1 ; Recursively update children
		__UIBB_UpdateClientArea($CHILD['CHILDREN'][$i], $CHILD, $i)
	Next
EndFunc   ;==>__UIBB_UpdateClientArea

Func __UIBB_UpdateCtrlClientArea(ByRef $CTRL, ByRef $PARENT, $idx)
	If Not __UIBB_IsBox($PARENT) Then
		ConsoleWrite('! ERROR: Cannot update client area of Ctrl... Ctrl is not in a Box... how?' & @CRLF)
		Return SetError(1, 0, 0)
	EndIf
	__UIBB_UpdateChildClientArea($CTRL, $PARENT, $idx)
	GUICtrlSetPos($CTRL['ID'], $CTRL['RECT'].x, $CTRL['RECT'].y, $CTRL['RECT'].w, $CTRL['RECT'].h)
EndFunc   ;==>__UIBB_UpdateCtrlClientArea

Func __UIBB_ComputeEffectiveScale(ByRef $PARENT)
	Local $iChildren = UBound($PARENT['CHILDREN'])
	If Not $iChildren Then Return

	; Default Total Scale
	Local $nScale = 0, $nScaleTotal = 0, $bNonBlocking = False
	For $i = 0 To $iChildren - 1 Step 1
		$bNonBlocking = $PARENT['CHILDREN'][$i]['NONBLOCKING']
		If Not $bNonBlocking Then
			$nScaleTotal += $PARENT['CHILDREN'][$i]['SCALE']
			$PARENT['CHILDREN'][$i]['EFFSCALE'] = $PARENT['CHILDREN'][$i]['SCALE']
		EndIf
	Next

	; Check if Scale does not fit MinMax requirements and fit accordingly
	Local $nScaleTotalNew = $nScaleTotal, $nParentWidth, $nParentHeight, $nMin, $nMax
	Local $nEffectiveWidth = $PARENT['RECT'].w - $PARENT['MARGIN'].l - $PARENT['MARGIN'].r
	Local $nEffectiveHeight = $PARENT['RECT'].h - $PARENT['MARGIN'].t - $PARENT['MARGIN'].b

	Switch $PARENT['TYPE']
		Case $UIBB_HBOX
			For $iter = 0 To $UIBB_MAXITER - 1 Step 1 ; Fit Iterations -> 2 is "ok", 3 is "good", 10 is "exact"
				$nParentWidth = $nEffectiveWidth / $nScaleTotalNew
				$nParentHeight = $nEffectiveHeight / $nScaleTotalNew
				$nScaleTotalNew = 0
				For $i = 0 To $iChildren - 1 Step 1
					$bNonBlocking = $PARENT['CHILDREN'][$i]['NONBLOCKING']
					If Not $bNonBlocking Then
					$nScale = $PARENT['CHILDREN'][$i]['EFFSCALE']
					$nMin = $PARENT['CHILDREN'][$i]['MINSIZE']
					$nMax = $PARENT['CHILDREN'][$i]['MAXSIZE']
					If $nScale * $nParentWidth < $nMin Then $nScale = $nMin / $nParentWidth
					If $nScale * $nParentWidth > $nMax Then $nScale = $nMax / $nParentWidth
					$PARENT['CHILDREN'][$i]['EFFSCALE'] = $nScale
					$nScaleTotalNew += $nScale
					EndIf
				Next
			Next
		Case $UIBB_VBOX
			For $iter = 0 To $UIBB_MAXITER - 1 Step 1 ; Fit Iterations -> 2 is "ok", 3 is "good", 10 is "exact"
				$nParentWidth = $PARENT['RECT'].w / $nScaleTotalNew
				$nParentHeight = $PARENT['RECT'].h / $nScaleTotalNew
				$nScaleTotalNew = 0
				For $i = 0 To $iChildren - 1 Step 1
					$nScale = $PARENT['CHILDREN'][$i]['EFFSCALE']
					$nMin = $PARENT['CHILDREN'][$i]['MINSIZE']
					$nMax = $PARENT['CHILDREN'][$i]['MAXSIZE']
					If $nScale * $nParentHeight < $nMin Then $nScale = $nMin / $nParentHeight
					If $nScale * $nParentHeight > $nMax Then $nScale = $nMax / $nParentHeight
					$PARENT['CHILDREN'][$i]['EFFSCALE'] = $nScale
					$nScaleTotalNew += $nScale
				Next
			Next
	EndSwitch
EndFunc   ;==>__UIBB_ComputeEffectiveScale

Func __UIBB_UpdateChildClientArea(ByRef $CHILD, ByRef $PARENT, $idx)
	Local $iNumChildren = UBound($PARENT['CHILDREN'])
	Local $nScaleTotal = 0, $nScaleUntilMe = 0, $nScaleCurrent = 0, $bNonBlocking = False
	For $i = 0 To UBound($PARENT['CHILDREN']) - 1 Step 1
		$bNonBlocking = $PARENT['CHILDREN'][$i]['NONBLOCKING']
		If Not $bNonBlocking Then
			$nScaleCurrent = $PARENT['CHILDREN'][$i]['EFFSCALE']
			$nScaleTotal += $nScaleCurrent
			If $i < $idx Then $nScaleUntilMe += $nScaleCurrent
		EndIf
	Next
	$CHILD['RECT'] = __UIBB_Rect($PARENT['RECT'])
	$CHILD['RECT'].x += $PARENT['MARGIN'].l
	$CHILD['RECT'].y += $PARENT['MARGIN'].t
	$CHILD['RECT'].w -= $PARENT['MARGIN'].l + $PARENT['MARGIN'].r
	$CHILD['RECT'].h -= $PARENT['MARGIN'].t + $PARENT['MARGIN'].b

	If $iNumChildren > 0 Then
		Switch $PARENT['TYPE']
			Case $UIBB_HBOX
				$CHILD['RECT'].x += $nScaleUntilMe / $nScaleTotal * $CHILD['RECT'].w
				$CHILD['RECT'].w *= $CHILD['EFFSCALE'] / $nScaleTotal
			Case $UIBB_VBOX
				$CHILD['RECT'].y += $nScaleUntilMe / $nScaleTotal * $CHILD['RECT'].h
				$CHILD['RECT'].h *= $CHILD['EFFSCALE'] / $nScaleTotal
		EndSwitch
	EndIf
EndFunc   ;==>__UIBB_UpdateChildClientArea

Func __UIBB_UpdateBoxClientArea(ByRef $CHILD, ByRef $PARENT, $idx)
	If MapExists($CHILD, 'GUI') Then
		Local $tRECT = DllStructCreate('int x1; int y1; int x2; int y2')
		DllCall('user32.dll', 'bool', 'GetClientRect', 'hwnd', $CHILD['GUI'], 'struct*', $tRECT)
		$CHILD['RECT'].x = $tRECT.x1
		$CHILD['RECT'].y = $tRECT.y1
		$CHILD['RECT'].w = $tRECT.x2 - $tRECT.x1
		$CHILD['RECT'].h = $tRECT.y2 - $tRECT.y1
		Return
	EndIf
	__UIBB_UpdateChildClientArea($CHILD, $PARENT, $idx)
EndFunc   ;==>__UIBB_UpdateBoxClientArea

Func __UIBB_IsBox(ByRef $CHILD)
	Return IsMap($CHILD) And (($CHILD['TYPE'] = $UIBB_HBOX) Or ($CHILD['TYPE'] = $UIBB_VBOX) Or ($CHILD['TYPE'] = $UIBB_EMPTY))
EndFunc   ;==>__UIBB_IsBox

Func __UIBB_IsCtrl(ByRef $CHILD)
	Return IsMap($CHILD) And IsFunc($CHILD['TYPE'])
EndFunc   ;==>__UIBB_IsCtrl

Func __UIBB_Rect($x = 0, $y = 0, $w = 0, $h = 0)
	Local $RECT[]
	If IsMap($x) Then
		$RECT.x = $x.x
		$RECT.y = $x.y
		$RECT.w = $x.w
		$RECT.h = $x.h
	Else
		$RECT.x = $x
		$RECT.y = $y
		$RECT.w = $w
		$RECT.h = $h
	EndIf
	Return $RECT
EndFunc   ;==>__UIBB_Rect

Func __UIBB_Margin($left = 0, $right = 0, $top = 0, $bot = 0)
	Local $MARGIN[]
	If IsMap($left) Then
		$MARGIN.l = $left.l
		$MARGIN.r = $left.r
		$MARGIN.t = $left.t
		$MARGIN.b = $left.b
	Else
		$MARGIN.l = $left
		$MARGIN.r = $right
		$MARGIN.t = $top
		$MARGIN.b = $bot
	EndIf

	Return $MARGIN
EndFunc   ;==>__UIBB_Rect

